/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */
import Foundation
import DeviceLib

public class DataConnector {
    
    private let humiditySensor: HumiditySensor
    private let HUMIDITY_ATTRIBUTE = "humidity"
    private let MAX_THRESHOLD_ATTRIBUTE = "maxThreshold"
    private let HUMIDITY_SENSOR_MODEL_URN: String = "urn:com:oracle:iot:device:humidity_sensor"
    private var dcd: DirectlyConnectedDevice?
    
    private var outputText: String = ""
    
    private var humidity: Int = 0
    
    private var maxMaxHygroThreshold: Int = 95
    private var minMaxHygroThreshold: Int = 80
    private var rememberMaxThreshold: Int  = 0
    
    private var stopped: Bool = false
    private var vd: VirtualDevice?
    private var firstOutput: Bool = true
    
    // Error message in case the initialization of the device failed.
    private var errorMessage = String()
    
    public init() throws {
        dcd = try DirectlyConnectedDevice(path: Provisioning.getProvisioningFilePath(),
                                          password: Provisioning.getProvisioningPassword())
        
        let persistentData = UserDefaults.standard
        if let hardwareId = persistentData.string(forKey: self.dcd!.getEndpointId()) {
            humiditySensor = HumiditySensor(id: hardwareId)
        } else {
            humiditySensor = HumiditySensor(id: UUID().uuidString)
            persistentData.set(humiditySensor.getHardwareId(),
                               forKey: self.dcd!.getEndpointId())
        }
   }
    
    internal func isDeviceInitialized() -> Bool {
        return self.dcd != nil ? true : false
    }
    
    internal func getErrorMessage() -> String {
        return errorMessage
    }
    
    public func getOutputText() -> String {
        return self.outputText
    }
    
    public func getMaxThreshold() -> Int {
        return humiditySensor.getMaxThreshold()
    }
    
    public func getHumidity() -> Int {
        return self.humidity
    }
    
    public func getEndpointId() -> String {
        return self.dcd!.getEndpointId()
    }
    
    public func increaseMaxThreshold() {
        let mt: Int = humiditySensor.getMaxThreshold() + 1
        if mt > self.maxMaxHygroThreshold {
            display(string: "Maximum of humidty maxThreshold (" +
                maxMaxHygroThreshold.description + ") reached!")
        } else {
            humiditySensor.setMaxThreshold(threshold: mt)
        }
    }
    
    public func decreaseMaxThreshold() {
        let mt: Int  = humiditySensor.getMaxThreshold() - 1
        if mt < self.minMaxHygroThreshold {
            display(string: "Minimum of humidty maxThreshold (" +
                minMaxHygroThreshold.description + ") reached!")
        } else {
            humiditySensor.setMaxThreshold(threshold: mt)
        }
    }
    
    public func stop() {
        self.stopped = true
        sleep(2)
        outputText = ""
    }
    
    public func unstop() {
        self.stopped = false
        self.firstOutput = true
    }
    
    public func close() {
        if(self.dcd != nil) {
            do {
                try dcd!.close()
            } catch {
                print("Issue while closing Directly Connected Device.")
            }
            dcd = nil
        }
    }
    
    deinit {
        if(self.dcd != nil) {
            do {
                try dcd!.close()
            } catch {
                print("Issue while closing Directly Connected Device.")
            }
            dcd = nil
        }
    }
    
    public func startSample() {
        self.createVirtualDevice(callback: { error in
            if (error == nil) {
                if self.rememberMaxThreshold != 0 {
                    self.humiditySensor.setMaxThreshold(threshold: self.rememberMaxThreshold)
                }
                let startHumidity: Int = self.humiditySensor.getHumidity()
                let maxThreshold: Int = self.humiditySensor.getMaxThreshold()
                self.rememberMaxThreshold = maxThreshold
                self.registerControlHandlers()
                self.updateVirtualDevice(humidity: startHumidity)
            } else {
                // An error has encounterd during device creation.
                print(error!)
            }
        })
    }
    
    func registerControlHandlers() {
        // Manage humidity sensor attribute changes of the device.
        vd!.setOnChange(callback: { event in
            let device: VirtualDevice = event.getVirtualDevice()
            // Update the device with the new value.
            var namedValue: NamedValue? = event.getNamedValue()
            var attribute: String
            var value: Int
            var str = self.getCurrentDateString() + " : "
                + device.getEndpointId() + " : onChange : "
            var first = true
            
            
            while (namedValue != nil) {
                if (!first) {
                    str = str + ","
                }
                else {
                    first = false
                }
                
                attribute = namedValue!.getName()
                value = namedValue!.getValue() as! Int
                
                if (attribute == "maxThreshold") {
                    str = str + "\"maxThreshold\"=" + String(value)
                    self.humiditySensor.setMaxThreshold(threshold: value)
                } else {
                    str = str + attribute + " not implemented"
                }
                
                namedValue = namedValue?.next()
            }
            
            self.display(string: str)
        })
        
        // Manage update errors for the humidity sensor device.
        vd!.setOnError(callback: { event in
            let device: VirtualDevice = event.getVirtualDevice()
            let errorMsg = event.getMessage()
            var str = self.getCurrentDateString() + " : onError : "
            str = str + device.getEndpointId() + errorMsg
            self.display(string: str)
        })
    }
    
    func updateData() {
        // A flag to make sure alerts are only sent when
        // crossing the threshold
        var alerted: Bool = false
        if (self.stopped != true) {
            self.humidity = self.humiditySensor.getHumidity()
            self.updateVirtualDevice(humidity: self.humidity)
            let threshold: Int = self.humiditySensor.getMaxThreshold()
            if (self.humidity < threshold) {
                alerted = false
            } else if (!alerted) {
                alerted = true;
                sendAlert(value: self.humidity)
            }
        }
    }
    
    func getCurrentDateString() -> String {
        let formatter = DateFormatter()
        formatter.timeStyle = .medium
        formatter.dateStyle = .medium
        let currentDateTime = Date()
        return formatter.string(from: currentDateTime)
    }
    
    func display(string: String) {
        outputText += "\n" + string
    }
    
    // Activate the humidity sensor device.
    func activateDevice(callback: @escaping (ClientError?) -> Void) {
        if !(self.dcd!.isActivated()) {
            do {
                try self.dcd!.activate(deviceModels: HUMIDITY_SENSOR_MODEL_URN,
                    callback: { endpointId, error in
                        callback(error)
                    })
            } catch {
                print(error)
            }
        } else {
            callback(nil)
        }
    }
    
    func createVirtualDevice(callback: @escaping (ClientError?) -> Void) {
        do {
            try self.dcd!.getDeviceModel(deviceModelUrn: self.HUMIDITY_SENSOR_MODEL_URN,
                                         callback: { (returnDeviceModel, error) in
                if error != nil {
                    callback(error)
                    return
                }
                if let returnObject = returnDeviceModel {
                    do {
                        self.vd = try self.dcd!.createVirtualDevice(deviceId: self.dcd!.getEndpointId(),
                                                                    deviceModel: returnObject)
                    } catch  {
                        print(error)
                        callback(error as? ClientError)
                        return
                    }
                }
                // set the default threshold value with the device model's
                // default value
                do {
                    self.humiditySensor.setMaxThreshold(
                        threshold: Int(truncating: try self.vd!.get(attributeName: self.MAX_THRESHOLD_ATTRIBUTE) as! NSNumber))
                } catch {
                    // Just use the hardcoded default max threshold
                    print(error)
                }
                callback(error)
            })
        } catch {
            print(error)
        }
    }

    // Update the value of the attribute "humidity"
    func updateVirtualDevice(humidity: Int) -> Void {
        do {
            let str: String = self.getCurrentDateString() + " : " +
                self.dcd!.getEndpointId() +
                " : Set : \"humidity\"=" + humidity.description

            if (self.firstOutput) {
                if let _vd = vd {
                    let _ = try _vd.update().set(attributeName: "humidity",
                                        attributeValue: humidity as AnyObject )
                    try _vd.finish()
                    self.firstOutput = false
                } else {
                    print("updateVirtualDevice: Still don't have virtual device")
                }
            } else {
                let _ = try vd!.set(attributeName: "humidity",
                                    attributeValue: humidity as AnyObject )
            }
            self.display(string: str)
        } catch {
            print("Failed to update virtual device")
        }
    }
    
    // Create and set an alert ":too_humid"
    func sendAlert(value: Int) -> Void {
        do {
            if let _vd = vd {
                let al:Alert = try _vd.createAlert(format: HUMIDITY_SENSOR_MODEL_URN + ":too_humid")
                try al.set(fieldName: HUMIDITY_ATTRIBUTE,
                           value: value as AnyObject).raise()
                let str = self.getCurrentDateString() + " : " +
                    self.dcd!.getEndpointId() + " : Alert : " +
                    "\"humidity\"=" + value.description
                self.display(string: str)
            } else {
                print("sendAlert: Still don't have virtual device")
            }
        } catch {
            print("Failed to send alert")
        }
    }
}

