/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */


import UIKit

@objc class DeviceDetailView: UIViewController, UITextViewDelegate {
    
    // Needed for the loops in the event handler button actions below.
    @objc var timer1: Timer?
    @objc var timer2: Timer?
    
    @objc var stopped:Bool = false
    
    //MARK: Properties
    
    @IBOutlet weak var headLineLabelOutlet: UILabel!
    @IBOutlet weak var sensorLabelOutlet: UILabel!
    @IBOutlet weak var hygroPicImageViewOutlet: UIImageView!
    @IBOutlet weak var idValueOutlet: UILabel!
    @IBOutlet weak var humidityLabelOutlet: UILabel!
    @IBOutlet weak var humidityValueOutlet: UILabel!
    @IBOutlet weak var maxThresholdLabelOutlet: UILabel!
    @IBOutlet weak var maxThresholdValueOutlet: UILabel!

    @IBOutlet weak var outputLabelOutlet: UILabel!
    @IBOutlet weak var outputScrollViewOutlet: UIScrollView!
    @IBOutlet weak var outputTextViewOutlet: UITextView!
    
    @IBOutlet weak var stopButtonOutlet: UIButton!
    @IBOutlet weak var resetButtonOutlet: UIButton!

    @IBOutlet var ViewControllerOutlet: UIView!
    
    override func viewDidLoad() {
        super.viewDidLoad()
        outputTextViewOutlet.delegate = self
    }
    
    private func updateGui(id: String?, humidity: String?,  maxThreshold: String?,
                           text: String? = nil) {
        DispatchQueue.main.async(execute: {
            self.idValueOutlet.text = id ?? "<id>"
            self.humidityValueOutlet.text = humidity ?? "<value> %"
            self.maxThresholdValueOutlet.text = maxThreshold ?? "<value> %"

            if let text = text {
                self.outputTextViewOutlet.text = text
                // Force text field to (invisibly) scroll until latest line is at bottom
                let bottom: NSRange = NSMakeRange(self.outputTextViewOutlet.text.count - 10, 1)
                self.outputTextViewOutlet.scrollRangeToVisible((bottom))
            } else {
                // assume its initialization, stop or reset
                self.outputTextViewOutlet.text = kNotAttached
            }
        })
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        // If we reach this screen going back is not an option.
        // There may also be problems provisioning again since the 
        // provisioning file will exist
        self.navigationController?.setNavigationBarHidden(true,
                                                          animated: animated)
        
        do {
            try self.dataConnector = DataConnector()
            DeviceDetailView._dataConnector = self.dataConnector
        } catch {
            print(error)
            ChooseProvisionerView.haveProvisioningError = true
            showError(controller: self,alertTitle: "Provisioning Failure",
                      message: kProvisioningFailure, actionTitle: "OK",
                      handler: { (action:UIAlertAction) -> Void in
                let _ = self.navigationController?.popToRootViewController(animated: false)
            })
        }
        
        if dataConnector != nil {
            updateGui(id: "<id>", humidity: "<value> %",
                      maxThreshold: "<value> %")
            if !dataConnector!.isDeviceInitialized() {
                showError(controller: self,
                          alertTitle: "Device Initialization Alert",
                          message: dataConnector!.getErrorMessage(),
                          actionTitle: "Exit App",
                          handler: { (action: UIAlertAction) -> Void in exit(0)})
                return
            } else {
                self.stopped = false
                self.dataConnector!.unstop()
            
                // Activate the humidity sensor device and start pushing data.
                self.dataConnector!.activateDevice(callback: { error in
                    // Create and manage the device.
                    self.dataConnector!.startSample()
                    // Loop for sample measurements
                    // Timers must be started on the main thread.
                    // This code is executed from the Alamofire response 
                    // queue. This queue has been changed to be NOT the main
                    // queue.
                    DispatchQueue.main.async(execute:{
                        self.timer1 = Timer.scheduledTimer(timeInterval: 5.0,
                               target: self,
                               selector: #selector(DeviceDetailView.updateData),
                               userInfo: nil, repeats: true)
                    // Loop to ask for new measurement values provided by the DCD sample class
                        self.timer2 = Timer.scheduledTimer(timeInterval: 2.0,
                               target: self,
                               selector: #selector(DeviceDetailView.askSensor),
                               userInfo: nil, repeats: true)
                    })
                })
            }
        }
    }
    
    override func viewDidAppear(_ animated: Bool) {
        if self.dataConnector != nil {
            self.navigationItem.setHidesBackButton(true, animated: false)
        }
        super.viewDidAppear(animated)
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    // Close the data connector and all used resources, device, tam and 
    // communication channel
    @objc func closeDataConnector() {
        self.dataConnector!.close()
    }
    
    // TODO: Hack to not crash when app is closed
    // Need to find out how this worked before the ChooserProvisioner screen
    // Since DeviceDetailView has never been the root screen
    static var _dataConnector: DataConnector!
    @objc static func closeDataConnector() {
        _dataConnector?.close()
    }

    var dataConnector: DataConnector? = nil

    @objc func askSensor() {
        if (!self.stopped) {
            updateGui(id: self.dataConnector!.getEndpointId(),
                      humidity: "\(self.dataConnector!.getHumidity()) %",
                maxThreshold: "\(self.dataConnector!.getMaxThreshold()) %",
                text: self.dataConnector!.getOutputText())
        }
    }

    @objc func updateData() {
        self.dataConnector!.updateData()
    }
    
    // MARK: Actions
    
    // Action for the button "Stop"
    @IBAction func onStopButton(_ sender: UIButton) {
        self.dataConnector!.stop()
        self.stopped = true
        // Stop the loops
        self.timer1?.invalidate()
        self.timer2?.invalidate()
        outputTextViewOutlet.text = kNotAttached
        updateGui(id: "<id>", humidity: "<value> %", maxThreshold: "<value> %")

        // TODO: This is only needed in the network and file provisioning cases
        // For bundled provisioning, the TrustStore.plist file and the provisioning
        // file will still exist
        // Save the last known provisioning file information
        saveSettingsFile(path:
            URL(fileURLWithPath: Provisioning.getProvisioningFilePath()).lastPathComponent,
            password: String(Provisioning.getProvisioningPassword()))
        exit(0)
    }

    // Action for the button "Reset"
    @IBAction func onResetButton(_ sender: UIButton) {
        self.dataConnector!.stop()
        self.stopped = true
        // Stop the loops
        self.timer1?.invalidate()
        self.timer2?.invalidate()
        outputTextViewOutlet.text = kNotAttached
        updateGui(id: "<id>", humidity: "<value> %", maxThreshold: "<value> %")
        // Allow reprovisioning
        removeLastSettingsFile()
        // This should be the choose provisioning screen
        let _ = self.navigationController?.popToRootViewController(animated: false)
        //self.performSegue(withIdentifier: "unwindToChooseProvisioner", sender: self)
    }
}

