/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */
import Foundation
import DeviceLib

public class DataConnector {
    
    let HUMIDITY_ATTRIBUTE = "humidity"
    let TEMPERATURE_ATTRIBUTE = "temp"
    let MAX_THRESHOLD_ATTRIBUTE = "maxThreshold"
    let MIN_THRESHOLD_ATTRIBUTE = "minThreshold"
    let UNIT_ATTRIBUTE = "unit"
    let TOO_HUMID_ALERT = "urn:com:oracle:iot:device:humidity_sensor:too_humid"
    let TOO_COLD_ALERT = "urn:com:oracle:iot:device:temperature_sensor:too_cold"
    let TOO_HOT_ALERT = "urn:com:oracle:iot:device:temperature_sensor:too_hot"
    
    static let HUMIDITY_SENSOR_MODEL_URN: String = "urn:com:oracle:iot:device:humidity_sensor"
    static let TEMPERATURE_SENSOR_MODEL_URN: String = "urn:com:oracle:iot:device:temperature_sensor"
    
    private var timer: Timer = Timer()

    // If this variable is true then the icd will be bound to this gateway
    // and not allowed to roam
    private var icdVirtDevTempRestricted: Bool = true //false

    private var icdVirtualDeviceTemperature: VirtualDevice?
    private var tempSensor: TemperatureSensor?
    private var tempSensorEndpointId: String = ""
    
    // If this variable is true then the icd will be bound to this gateway
    // and not allowed to roam
    private var icdVirtDevHumidRestricted: Bool = true //false
    private var icdVirtualDeviceHumidity: VirtualDevice?
    private var humiditySensor: HumiditySensor?
    private var humiditySensorEndpointId: String = ""
    
    private var outputText: String = ""
    
    // Set when the virtualdevice is created.
    // Used to reset thresholds on power on
    private var defaultMinTempThreshold: Int?
    private var defaultMaxTempThreshold: Int?
    
    private var maxMaxThermoThreshold: Int = 70
    private var minMaxThermoThreshold: Int = 55
    private var maxMaxHygroThreshold: Int = 95
    private var minMaxHygroThreshold: Int = 80
    
    private var prevMinTemp : Double = 0.0
    private var prevMaxTemp : Double = 0.0
    private var minTemp : Double = 0.0
    private var maxTemp : Double = 0.0
    private var temperature : Double = 0.0
    private var humidity : Int = 0
    
    private var stopped: Bool = false
    private var firstMsg: Bool = true
    private var tempWasOff: Bool = true
    private var changedTemperatureThreshold: Bool = false
    private var changedHumidityThreshold: Bool = false
    
    var gd: GatewayDevice?
    
    var humAlert:Alert?
    var tempColdAlert:Alert?
    var tempHotAlert:Alert?
    
    // Error message in case the initialization of the device failed.
    private var errorMessage = String()

    public init() throws {
        gd = try GatewayDevice(path: Provisioning.getProvisioningFilePath(),
                            password: Provisioning.getProvisioningPassword())
    }
    
    internal func isDeviceInitialized() -> Bool {
        return self.gd != nil ? true : false
    }
    
    internal func getErrorMessage() -> String {
        return errorMessage
    }

    
    public func getOutputText() -> String {
        return self.outputText
    }
    
    public func getMaxThreshold() -> Int {
        return humiditySensor!.getMaxThreshold()
    }
    
    public func getThermoMaxThreshold() -> Int {
        return tempSensor!.getMaxThreshold()
    }
    
    public func getThermoMinThreshold() -> Int {
        return tempSensor!.getMinThreshold()
    }
    
    public func getEndpointId() -> String {
        return self.gd!.getEndpointId()
    }
    
    public func getHumidity() -> Int {
        return self.humidity
    }
    
    public func getTemperature() -> Double {
        return self.temperature
    }
    
    public func getMinTemperature() -> Double {
        return tempSensor!.getMinTemp()
    }
    
    public func getMaxTemperature() -> Double {
        return tempSensor!.getMaxTemp()
    }
    
    public func getStartTime() -> Int64 {
        return tempSensor!.getStartTime()
    }
    
    public func getEndpointIdH() -> String {
        return self.humiditySensorEndpointId
    }
    
    public func getEndpointIdT() -> String {
        return self.tempSensorEndpointId
    }
    
    public func increaseMaxThreshold() {
        let mt: Int = humiditySensor!.getMaxThreshold() + 1
        if mt > self.maxMaxHygroThreshold {
            display(string: "Maximum of humidty maxThreshold (" +
                maxMaxHygroThreshold.description + ") reached!")
        } else {
            humiditySensor!.setMaxThreshold(threshold: mt)
            changedHumidityThreshold = true
        }
    }
    
    public func increaseThermoMaxThreshold() {
        let mt: Int = tempSensor!.getMaxThreshold() + 1
        if mt > self.maxMaxThermoThreshold {
            display(string: "Maximum of temperature maxThreshold (" +
                maxMaxThermoThreshold.description + ") reached!")
        } else {
            tempSensor!.setMaxThreshold(threshold: mt)
            changedTemperatureThreshold = true
        }
    }
    
    public func decreaseMaxThreshold() {
        let mt: Int  = humiditySensor!.getMaxThreshold() - 1
        if mt < self.minMaxHygroThreshold {
            display(string: "Minimum of humidty maxThreshold (" +
                minMaxHygroThreshold.description + ") reached!")
        } else {
            humiditySensor!.setMaxThreshold(threshold: mt)
            changedHumidityThreshold = true
        }
    }
    
    public func decreaseThermoMaxThreshold() {
        let mt: Int  = tempSensor!.getMaxThreshold() - 1
        if mt < self.minMaxThermoThreshold {
            display(string: "Minimum of temperature maxThreshold (" +
                minMaxThermoThreshold.description  + ") reached!")
        } else {
            tempSensor!.setMaxThreshold(threshold: mt)
            changedTemperatureThreshold = true
        }
    }
    
    public func stop() {
        self.stopped = true
        sleep(2)
        outputText = ""
    }
    
    public func unstop() {
        self.stopped = false
        self.firstMsg = true
    }
    
    public func close() {
        if(self.gd != nil) {
            do {
                try gd!.close()
            } catch {
                print("Issue while closing Gateway device.")
            }
            gd = nil
        }
    }
    
    func getCurrentDateString() -> String {
        let formatter = DateFormatter()
        formatter.timeStyle = .medium
        formatter.dateStyle = .medium
        let currentDateTime = Date()
        return formatter.string(from: currentDateTime)
    }
    
    func display(string: String) {
        outputText += "\n" + string
    }
    
    // Prepare the gateway device.
    func pgd(callback: @escaping (ClientError?) -> Void) {
        prepareGatewayDevice(callback: { error in
            if (error == nil) {
                self.humiditySensor = HumiditySensor(id: self.gd!.getEndpointId() + "_Sample_HS")
                self.tempSensor = TemperatureSensor(id: self.gd!.getEndpointId() + "_Sample_TS")
            } else {
                print(error!)
            }
            callback(error)
        })
    }
    
    // Activate the gateway device.
    func prepareGatewayDevice(callback: @escaping (ClientError?) -> Void) {
        if !(gd!.isActivated()) {
            do {
                try gd!.activate(callback: { endpointId, error in
                    callback(error)
                })
            } catch {
                print(error)
            }
        } else {
            callback(nil)
        }
    }
    
    // Register indirectly connectde devices.
    func registerIndirectlyConnectedDevices(callback: @escaping (ClientError?) -> Void) {
        var tempMetaData: [String : String] = [ : ]
        tempMetaData[GatewayDevice.MANUFACTURER] = tempSensor!.getManufacturer()
        tempMetaData[GatewayDevice.MODEL_NUMBER] = tempSensor!.getModelNumber()
        tempMetaData[GatewayDevice.SERIAL_NUMBER] = tempSensor!.getSerialNumber()
        
        var humidityMetaData: [String : String] = [ : ]
        humidityMetaData[GatewayDevice.MANUFACTURER] = humiditySensor!.getManufacturer()
        humidityMetaData[GatewayDevice.MODEL_NUMBER] = humiditySensor!.getModelNumber()
        humidityMetaData[GatewayDevice.SERIAL_NUMBER] = humiditySensor!.getSerialNumber()
        
        do {
            try self.gd!.registerDevice(restricted: self.icdVirtDevTempRestricted,
            hardwareId: self.tempSensor!.getHardwareId(),
                metaData: tempMetaData,
                deviceModels: DataConnector.TEMPERATURE_SENSOR_MODEL_URN,
                callback: { endpointIdT, rderrorT in
                    if (rderrorT != nil) {
                        callback(rderrorT)
                        return
                    }
                    self.tempSensorEndpointId = endpointIdT!
                    self.display(string: "\nCreated temperature sensor " + self.tempSensorEndpointId)
                    do {
                        try self.gd!.registerDevice(restricted: self.icdVirtDevHumidRestricted,
                            hardwareId: self.humiditySensor!.getHardwareId(),
                            metaData: humidityMetaData,
                            deviceModels: DataConnector.HUMIDITY_SENSOR_MODEL_URN,
                            callback: { endpointIdH, rderrorH in
                                if (rderrorH != nil) {
                                    callback(rderrorH)
                                }
                                self.humiditySensorEndpointId = endpointIdH!
                                self.display(string: "\nCreated humidity sensor " + self.humiditySensorEndpointId)
                                callback(nil)
                        })
                    } catch {
                        print (error)
                    }
                })
        } catch  {
            print (error)
        }
    }
    
    // Create virtual devices for humidity and temperature sensors.
    func createVirtualDevices(callback: @escaping (ClientError?) -> Void) {
        do {
            try self.gd!.getDeviceModel(deviceModelUrn: DataConnector.TEMPERATURE_SENSOR_MODEL_URN,
                                        callback: { (returnDeviceModelT, error) in
                if (error != nil) {
                    callback(error)
                }
                if let returnObjectT = returnDeviceModelT {
                    do {
                        self.icdVirtualDeviceTemperature = try self.gd!.createVirtualDevice(deviceId: self.tempSensorEndpointId, deviceModel: returnObjectT)
                        // Set the sensor's default max and min values to the device
                        // model's default values
                        do {
                            self.tempSensor?.setMaxThreshold(threshold:
                                Int(truncating: try self.icdVirtualDeviceTemperature?.get(
                                    attributeName: self.MAX_THRESHOLD_ATTRIBUTE) as! NSNumber))
                            self.defaultMaxTempThreshold = self.tempSensor?.getMaxThreshold()
                            self.tempSensor?.setMinThreshold(threshold:
                                Int(truncating: try self.icdVirtualDeviceTemperature?.get(
                                    attributeName: self.MIN_THRESHOLD_ATTRIBUTE) as! NSNumber))
                            self.defaultMinTempThreshold = self.tempSensor?.getMinThreshold()
                        } catch {
                            print(error)
                        }
                        try self.tempHotAlert = self.icdVirtualDeviceTemperature?.createAlert(format: self.TOO_HOT_ALERT)
                        try self.tempColdAlert = self.icdVirtualDeviceTemperature?.createAlert(format: self.TOO_COLD_ALERT)

                        try self.gd!.getDeviceModel(deviceModelUrn: DataConnector.HUMIDITY_SENSOR_MODEL_URN, callback: { (returnDeviceModelH, error) in
                            if let returnObjectH = returnDeviceModelH {
                                do {
                                    self.icdVirtualDeviceHumidity = try self.gd!.createVirtualDevice(deviceId: self.humiditySensorEndpointId, deviceModel: returnObjectH)
                                    try self.humAlert = self.icdVirtualDeviceHumidity?.createAlert(format: self.TOO_HUMID_ALERT)
                                    do {
                                        self.humiditySensor?.setMaxThreshold(threshold:
                                            Int(truncating: try self.icdVirtualDeviceHumidity?.get(
                                                attributeName: self.MAX_THRESHOLD_ATTRIBUTE) as! NSNumber))
                                    }
                                } catch {
                                    print(error)
                                }
                            }
                            callback(error)
                        })
                    } catch {
                        print(error)
                    }
                }
            })
        } catch {
            print (error)
        }
    }
    
    func startLoop() {
        // Loop for sample measurements
        // Timers must be started on the main thread.
        // This code is executed from the Alamofire response
        // queue. This queue has been changed to be NOT the main
        // queue.
        DispatchQueue.main.async(execute: {
            self.timer = Timer.scheduledTimer(timeInterval: 5, target: self,
                        selector: #selector(DataConnector.updateVirtualDevices),
                        userInfo: nil, repeats: true)
        })
    }

    // Create virtual devices (humidity and temperature) and start attribute 
    // updates.
    func startSample() {
        self.registerIndirectlyConnectedDevices(callback: { error in
            if (error == nil) {
                self.createVirtualDevices(callback: { error in
                    if (error == nil) {
                        self.registerControlHandlers()
                        self.startLoop()
                    } else {
                        print(error!)
                    }
                })
            } else {
                print(error!)
            }
        })
    }
    
    func registerControlHandlers() {
        do {
            // Manage humidity sensor attribute changes of the indirectly 
            // connected device.
            try icdVirtualDeviceHumidity!.setOnChange(attributeName: "maxThreshold",
                                                      callback: { event in
                let device: VirtualDevice = event.getVirtualDevice()
                let namedValue: NamedValue = event.getNamedValue()
                let value: Int = namedValue.getValue() as! Int
                let str = self.getCurrentDateString() + " : "
                    + device.getEndpointId() +
                    " : onChange : \"maxThreshold\"=\(value)"
                self.humiditySensor!.setMaxThreshold(threshold: value)
                self.display(string: str)
            })
        } catch {
            print("Error registering onChange handler")
        }
        
        // Manage update errors for the humidity sensor indirectly connected 
        // device.
        icdVirtualDeviceHumidity!.setOnError(callback: { event in
            let device: VirtualDevice = event.getVirtualDevice()
            let errorMsg = event.getMessage()
            let str = self.getCurrentDateString() + " : onError : " +
                device.getEndpointId() + errorMsg
            self.display(string: str)
        })
        
        // Manage temperature sensor attribute changes of the indirectly 
        // connected device.
        icdVirtualDeviceTemperature!.setOnChange(callback: { event in
            let device: VirtualDevice = event.getVirtualDevice()
            // Update the indirectly connected device with the new value.
            var namedValue: NamedValue? = event.getNamedValue()
            var attribute: String
            var value: Int
            var str = self.getCurrentDateString()  + " : "
                + device.getEndpointId() + " : onChange : "
            var first = true
            
            while (namedValue != nil) {
                if (!first) {
                    str = str + ","
                }
                else {
                    first = false
                }
                
                attribute = namedValue!.getName()
                value = namedValue!.getValue() as! Int
                
                if (attribute == "maxThreshold") {
                    str = str + "\"maxThreshold\"=\(value)"
                    self.tempSensor!.setMaxThreshold(threshold: value)
                } else if (attribute == "minThreshold") {
                    str = str + "\"minThreshold\"=\(value)"
                    self.tempSensor!.setMinThreshold(threshold: value)
                } else {
                    str = str + attribute + " not implemented"
                }
                namedValue = namedValue?.next()
            }
            self.display(string: str)
        })
        
        // Manage update errors for the temperature sensor indirectly connected 
        // device.
        icdVirtualDeviceTemperature!.setOnError(callback: { event in
            let device: VirtualDevice = event.getVirtualDevice()
            let errorMsg = event.getMessage()
            var str = self.getCurrentDateString()  + " : onError : "
            str = str + device.getEndpointId() + errorMsg
            self.display(string: str)
        })
        
        // Handle actions:
        do {
            try icdVirtualDeviceTemperature!.setOnCall(actionName: "power",
                                        callback: { (virtualDevice, data) in
                let on = data as! Bool
                let str = self.getCurrentDateString()  + " : " +
                    virtualDevice.getEndpointId() +
                    " : Call : \"power\"= \(on)"

                self.display(string: str)
                let isOn = self.tempSensor?.isPoweredOn()
                if (on != isOn) {
                    self.tempSensor?.power(on: on)
                }
            })

            try icdVirtualDeviceTemperature!.setOnCall(actionName: "reset",
                                        callback: { (virtualDevice, data) in
                var str = self.getCurrentDateString()  + " : " +
                    virtualDevice.getEndpointId()
                self.tempSensor?.reset()
                str = str + " : Call : reset : minTemp: " +
                    String(format: "%.2f", (self.tempSensor?.getMinTemp())!) +
                    ", maxTemp: " +
                    String(format: "%.2f", (self.tempSensor?.getMaxTemp())!)

                self.display(string: str)
            })
        } catch {
            print("Error registering action handlers")
        }
    }
    
    // Update the humidity and temperature sensor virtual devices attributes.
    @objc private func updateVirtualDevices() {
        if (!self.stopped) {
            var humAlerted:Bool = false
            var hotAlerted:Bool = false
            var coldAlerted:Bool = false
            let first:Bool = self.firstMsg
            
            if first {
                self.firstMsg = false
            }
            
            do {
                //Update Humidity sensor
                
                humidity = humiditySensor!.getHumidity()
                let _ = try self.icdVirtualDeviceHumidity!.update().set(
                                    attributeName: "humidity",
                                    attributeValue: humidity as AnyObject)
                var messageH = self.getCurrentDateString() + " : " +
                    humiditySensorEndpointId + " : Set : \"humidity\"=" +
                    humidity.description
                let humidityThreshold = self.humiditySensor!.getMaxThreshold()
                
                if changedHumidityThreshold {
                    changedHumidityThreshold = false
                    let _ = try self.icdVirtualDeviceHumidity!.set(
                            attributeName: "maxThreshold",
                            attributeValue: humidityThreshold as AnyObject)
                    messageH = messageH + " : \"maxThreshold\"=\(humidityThreshold)"
                }

                try self.icdVirtualDeviceHumidity!.finish()
                
                display(string: messageH)
                
                let newValueH: Int = try self.icdVirtualDeviceHumidity!.get(
                                            attributeName: "humidity") as! Int
                let thresholdH: Int = self.humiditySensor!.getMaxThreshold()
                
                if newValueH < thresholdH {
                    humAlerted = false
                } else if (!humAlerted) {
                    humAlerted = true
                    display(string: " : Alert : humidity = \(newValueH)")
                    let _ = try self.humAlert!.set(fieldName: HUMIDITY_ATTRIBUTE,
                                                   value: newValueH as AnyObject)
                    try self.humAlert!.raise()
                }
                
                
                //Update Temperature sensor if on
                
                if (!tempSensor!.isPoweredOn()) {
                    self.tempWasOff = true
                    return
                }
                
                prevMinTemp = tempSensor!.getMinTemp()
                prevMaxTemp = tempSensor!.getMaxTemp()
                temperature = tempSensor!.getTemp()
                minTemp = tempSensor!.getMinTemp()
                maxTemp = tempSensor!.getMaxTemp()
                
                let _ = try self.icdVirtualDeviceTemperature!.update().set(
                                    attributeName: "temp",
                                    attributeValue: temperature as AnyObject)
                
                var messageT = self.getCurrentDateString() + " : " +
                    tempSensorEndpointId + " : Set : \"temp\"=" +
                    temperature.description

                if (first || self.tempWasOff) {
                    if (self.tempWasOff) {
                        self.tempWasOff = false
                    }
                    
                    let _ = try self.icdVirtualDeviceTemperature!
                        .set(attributeName: "unit",
                             attributeValue: tempSensor!.getUnit() as AnyObject)
                        .set(attributeName: "minTemp",
                             attributeValue: tempSensor!.getMinTemp() as AnyObject)
                        .set(attributeName: "maxTemp",
                             attributeValue: tempSensor!.getMaxTemp() as AnyObject)
                        .set(attributeName: "startTime",
                             attributeValue: tempSensor!.getStartTime() as AnyObject)
                    
                    messageT = messageT + " : \"unit\"=\(tempSensor!.getUnit())"
                    messageT = messageT + " : \"minTemp\"=\(tempSensor!.getMinTemp())"
                    messageT = messageT + " : \"maxTemp\"=\(tempSensor!.getMaxTemp())"
                    
                    let minThreshold = tempSensor!.getMinThreshold()
                    let maxThreshold = tempSensor!.getMaxThreshold()
                    
                    // If minThreshold is not the default, include it

                    if minThreshold != defaultMinTempThreshold! {
                        let _ = try self.icdVirtualDeviceTemperature!.set(
                                    attributeName: "minThreshold",
                                    attributeValue: minThreshold as AnyObject)
                        messageT = messageT + " : \"minThreshold\"=\(tempSensor!.getMinThreshold())"
                    }
                    
                    // If maxThreshold is not the default, include it
                    if maxThreshold != defaultMaxTempThreshold! {
                        let _ = try self.icdVirtualDeviceTemperature!.set(
                                    attributeName: "maxThreshold",
                                    attributeValue: maxThreshold as AnyObject)
                        messageT = messageT + " : \"maxThreshold\"=\(tempSensor!.getMaxThreshold())"
                    }
                    
                    try self.icdVirtualDeviceTemperature!.finish()
 
                } else {
                    if (prevMinTemp != minTemp) {
                        let _ = try self.icdVirtualDeviceTemperature!.set(
                                    attributeName: "minTemp",
                                    attributeValue: minTemp as AnyObject)
                        messageT = messageT + " : \"minTemp\"=\(tempSensor!.getMinTemp())"
                    }
                    if (prevMaxTemp != maxTemp) {
                        let _ = try self.icdVirtualDeviceTemperature!.set(
                                    attributeName: "maxTemp",
                                    attributeValue: maxTemp as AnyObject)
                        messageT = messageT + " : \"maxTemp\"=\(tempSensor!.getMaxTemp())"
                    }
                    if changedTemperatureThreshold {
                        changedTemperatureThreshold = false
                        let _ = try self.icdVirtualDeviceTemperature!.set(
                                    attributeName: "maxThreshold",
                                    attributeValue: tempSensor!.getMaxThreshold() as AnyObject)
                        messageT = messageT + " : \"maxThreshold\"=\(tempSensor!.getMaxThreshold())"
                    }
                    
                    try self.icdVirtualDeviceTemperature!.finish()
                }
                
                display(string: messageT)
                
                let newValueT: Double = try self.icdVirtualDeviceTemperature!.get(
                                            attributeName: "temp") as! Double
                
                let maxThresholdT: Int = self.tempSensor!.getMaxThreshold()
                let minThresholdT: Int = self.tempSensor!.getMinThreshold()
                let unitT: String = self.tempSensor!.getUnit()
                
                if Int(newValueT) > minThresholdT {
                    coldAlerted = false
                } else if (!coldAlerted) {
                    coldAlerted = true
                    display(string: " : Alert : temperature = \(newValueT)")
                    let _ = try self.tempColdAlert!.set(fieldName: TEMPERATURE_ATTRIBUTE,
                                                        value: NSNumber(value: newValueT))
                    let _ = try self.tempColdAlert!.set(fieldName: UNIT_ATTRIBUTE,
                                                        value: unitT as AnyObject)
                    let _ = try self.tempColdAlert!.set(fieldName: MIN_THRESHOLD_ATTRIBUTE,
                                                        value: minThresholdT as AnyObject)
                    try self.tempColdAlert!.raise()
                }
                
                if Int(newValueT) < maxThresholdT {
                    hotAlerted = false
                } else if (!hotAlerted) {
                    hotAlerted = true
                    display(string: " : Alert : temperature = \(newValueT)")
                    let _ = try self.tempHotAlert!.set(fieldName: TEMPERATURE_ATTRIBUTE,
                                                       value: NSNumber(value: newValueT))
                    let _ = try self.tempHotAlert!.set(fieldName: UNIT_ATTRIBUTE,
                                                       value: unitT as AnyObject)
                    let _ = try self.tempHotAlert!.set(fieldName: MAX_THRESHOLD_ATTRIBUTE,
                                                       value: maxThresholdT as AnyObject)
                    try self.tempHotAlert!.raise()
                }
            } catch {
                print("Failed updating virtual devices.")
            }
        }
    }
}

