/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import QuartzCore

public class TemperatureSensor {
    public let START_TEMP: Double = 20
    
    private var x: Int = 0
    private var amplitude: Double = 0
    private var power: Bool = false
    private var startTime: Int64 = 0
    private var minTemp: Double = 0.0
    private var maxTemp: Double = 0.0
    private var setPoint: Double = 0
    private var minThreshold: Int
    private var maxThreshold: Int
    private var hardwareId: String
    private var persistentData : UserDefaults
    
    public init(id: String) {
        hardwareId = id
        
        persistentData = UserDefaults.standard
        if let min = persistentData.object(forKey: hardwareId + "minThreshold") {
            minThreshold = min as! Int
        } else {
            minThreshold = 0
            persistentData.set(minThreshold, forKey: hardwareId + "minThreshold")
        }
        if let max = persistentData.object(forKey: hardwareId + "maxThreshold") {
            maxThreshold = max as! Int
        } else {
            maxThreshold = 70
            persistentData.set(maxThreshold, forKey: hardwareId + "maxThreshold")
        }
        
        reset()
        power(on: true)
    }
    
    public func power(on: Bool) {
        if on {
            if self.power {
                return
            }
            
            self.startTime = Int64(Date().timeIntervalSince1970)
            self.power = true
            reset()
            return
        }
        self.power = false
    }
    
    public func isPoweredOn() -> Bool {
        return self.power
    }
    
    public func reset() {
        self.minTemp = 0.0
        self.maxTemp = 0.0
        self.setPoint = Double(self.maxThreshold) * 0.9
        self.amplitude = Double(self.maxThreshold) - self.setPoint + 1
    }
    
    /**
     * Get the current temperature value.
     * The simulated readings will fluctuate +/- 2 degrees.
     *
     * - returns: temperature value
     */
    public func getTemp() -> Double {
        let delta: Double = self.amplitude * Double(sin(Double(self.x) * Double.pi / 180))
        self.x = self.x + 30
        // Normal temp sensors don't have readings such as 21.890890309842234
        let l: UInt64 = UInt64((self.setPoint + delta) * 100)
        let temp: Double = Double(l) / 100
        
        if minTemp != 0.0 {
            if temp < minTemp {
                minTemp = temp
            }
        } else {
            minTemp = temp
        }
        if maxTemp != 0.0 {
            if temp > maxTemp {
                maxTemp = temp
            }
        } else {
            maxTemp = temp
        }
        return temp
    }
    
    public func getUnit() -> String {
        return "Celsius"
    }
    
    public func getMinTemp() -> Double {
        return self.minTemp
    }
    
    public func getMaxTemp() -> Double {
        return self.maxTemp
    }
    
    public func getMinThreshold() -> Int {
        return self.minThreshold
    }
    
    public func getMaxThreshold() -> Int {
        return self.maxThreshold
    }
    
    public func setMinThreshold(threshold: Int) {
        self.minThreshold = threshold
        persistentData.set(threshold, forKey: hardwareId + "minThreshold")
    }
    
    public func setMaxThreshold(threshold: Int) {
        self.maxThreshold = threshold
        persistentData.set(threshold, forKey: hardwareId + "maxThreshold")
        self.setPoint = Double(self.maxThreshold) * 0.9
        self.amplitude = Double(self.maxThreshold) - self.setPoint + 1
    }
    
    public func getStartTime() -> Int64 {
        return self.startTime
    }
    
    public func getManufacturer() -> String {
        return "Sample"
    }
    
    public func getModelNumber() -> String {
        return "MN-" + self.hardwareId
    }
    
    public func getSerialNumber() -> String {
        return "SN-" + self.hardwareId
    }
    
    public func getHardwareId() -> String {
        return self.hardwareId
    }
}
